SYSTEM_PROMPT = """Your input fields are:
{input_fields}

Your output fields are:
{output_fields}

All interactions will be structured in the following way, with the appropriate values filled in.

{interaction_format}

[[ ## completed ## ]]

In adhering to this structure, your objective is:
    {objective}."""
